unit UPascalParser;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  StdCtrls;

type

  { TTokenStream }

  TTokenStream = class
    Items: TStringList;
    constructor Create;
    destructor Destroy; override;
  end;

  { TPascalTokenizer }

  TPascalTokenizer = class
    Tokens: TTokenStream;
    procedure Tokenize(SourceCode: string);
    constructor Create;
    destructor Destroy; override;
  end;

implementation

{ TTokenStream }

constructor TTokenStream.Create;
begin
  Items := TStringList.Create;
end;

destructor TTokenStream.Destroy;
begin
  Items.Destroy;
  inherited Destroy;
end;

{ TPascalTokenizer }

procedure TPascalTokenizer.Tokenize(SourceCode: string);
type
  TTokenState = (ttNone, ttIdentifier, ttNumber, ttString, ttSingleLineComment,
    ttBlockComment, ttBlockComment2, ttSpecialSymbol, ttWhiteSpace,
    ttShortOperator, ttLongOperator, ttStringEnd);
var
  I: Integer;
  State: TTokenState;
  TokenStart: Integer;
begin
  Tokens.Items.Clear;
  I := 1;
  SourceCode := SourceCode + ' '; // End symbol test workaround
  State := ttNone;
  while I <= Length(SourceCode) do begin
    case State of
      ttNone: begin
        if ((SourceCode[I] >= 'A') and (SourceCode[I] <= 'Z')) or
          ((SourceCode[I] >= 'a') and (SourceCode[I] <= 'z')) then begin
            State := ttIdentifier;
            TokenStart := I;
        end else
        if SourceCode[I] <= ' ' then begin
          State := ttWhiteSpace;
          TokenStart := I;
        end else
        if (SourceCode[I] = '*') or (SourceCode[I] = '+') or
        (SourceCode[I] = '-') or (SourceCode[I] = '/') or
        (SourceCode[I] = '=') or (SourceCode[I] = '<') or
        (SourceCode[I] = '>') or (SourceCode[I] = '^') or
        (SourceCode[I] = ':') then begin
          State := ttShortOperator;
          TokenStart := I;
        end else
        if ((SourceCode[I] >= '0') and (SourceCode[I] <= '9')) or
        (SourceCode[I] = '$') then begin
          State := ttNumber;
          TokenStart := I;
        end else
        if (SourceCode[I] = '''') then begin
          State := ttString;
          TokenStart := I;
        end else
        if (SourceCode[I] = ')') or (SourceCode[I] = '(') or
        (SourceCode[I] = '.') or (SourceCode[I] = ';') then begin
          State := ttSpecialSymbol;
          TokenStart := I;
        end;
      end;
      ttIdentifier: begin
        if not (((SourceCode[I] >= 'A') and (SourceCode[I] <= 'Z')) or
          ((SourceCode[I] >= 'a') and (SourceCode[I] <= 'z')) or
          ((SourceCode[I] >= '0') and (SourceCode[I] <= '9'))) then begin
            Tokens.Items.Add(Copy(SourceCode, TokenStart, I - TokenStart));
            State := ttNone;
            Dec(I);
          end;
      end;
      ttWhiteSpace: begin
        if not (SourceCode[I] <= ' ') then begin
          //Tokens.Items.Add('');
          State := ttNone;
          Dec(I);
        end;
      end;
      ttShortOperator: begin
        if not (SourceCode[I] = '=') then begin
          Tokens.Items.Add(Copy(SourceCode, TokenStart, I - TokenStart));
          State := ttNone;
          Dec(I);
        end else State := ttLongOperator;
      end;
      ttLongOperator: begin
        Tokens.Items.Add(Copy(SourceCode, TokenStart, I - TokenStart));
        State := ttNone;
        Dec(I);
      end;
      ttNumber: begin
        if not (((SourceCode[I] >= '0') and (SourceCode[I] <= '9'))) then begin
            Tokens.Items.Add(Copy(SourceCode, TokenStart, I - TokenStart));
            State := ttNone;
            Dec(I);
          end;
      end;
      ttSpecialSymbol: begin
        Tokens.Items.Add(Copy(SourceCode, TokenStart, I - TokenStart));
        State := ttNone;
        Dec(I);
      end;
      ttString: begin
        if (SourceCode[I] = '''') then begin
          State := ttStringEnd;
        end;
      end;
      ttStringEnd: begin
        if (SourceCode[I] = '''') then begin
          State := ttString;
        end else begin
          Tokens.Items.Add(Copy(SourceCode, TokenStart, I - TokenStart));
          State := ttNone;
          Dec(I);
        end;
      end;
    end;
    Inc(I);
  end;
end;

constructor TPascalTokenizer.Create;
begin
  Tokens := TTokenStream.Create;
end;

destructor TPascalTokenizer.Destroy;
begin
  Tokens.Destroy;
  inherited Destroy;
end;

initialization
  {$I UMainForm.lrs}

end.

